# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from .offset_manager import OffsetManagerBase
from kafka_utils.kafka_consumer_manager.util import get_kafka_group_reader


class ListGroups(OffsetManagerBase):

    @classmethod
    def setup_subparser(cls, subparsers):
        parser_list_groups = subparsers.add_parser(
            "list_groups",
            description="List consumer groups.",
            add_help=False,
        )
        parser_list_groups.set_defaults(command=cls.run)

    @classmethod
    def get_kafka_groups(cls, cluster_config, use_admin_client=False):
        '''Get the group_id of groups committed into Kafka.'''
        kafka_group_reader = get_kafka_group_reader(cluster_config, use_admin_client)
        groups_and_topics = kafka_group_reader.read_groups(list_only=True)
        return list(groups_and_topics.keys())

    @classmethod
    def print_groups(cls, groups, cluster_config):
        print("Consumer Groups:")
        for groupid in groups:
            print(f"\t{groupid}")
        print(
            "{num_groups} groups found for cluster {cluster_name} "
            "of type {cluster_type}".format(
                num_groups=len(groups),
                cluster_name=cluster_config.name,
                cluster_type=cluster_config.type,
            ),
        )

    @classmethod
    def run(cls, args, cluster_config):
        groups = set()
        kafka_groups = cls.get_kafka_groups(
            cluster_config,
            args.use_admin_client,
        )
        if kafka_groups:
            groups.update(kafka_groups)

        cls.print_groups(groups, cluster_config)



# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import sys
from collections import defaultdict

from kafka.admin import KafkaAdminClient
from kafka.consumer import KafkaConsumer
from kafka.structs import OffsetAndMetadata
from kafka.structs import TopicPartition
from kafka.util import read_short_string
from kafka.util import relative_unpack
from kazoo.exceptions import NodeExistsError

from kafka_utils.util.client import KafkaToolClient
from kafka_utils.util.error import UnknownTopic
from kafka_utils.util.metadata import get_topic_partition_metadata
from kafka_utils.util.offsets import get_topics_watermarks


CONSUMER_OFFSET_TOPIC = '__consumer_offsets'


def preprocess_topics(source_groupid, source_topics, dest_groupid, topics_dest_group):
    """Pre-process the topics in source and destination group for duplicates."""
    # Is the new consumer already subscribed to any of these topics?
    common_topics = [topic for topic in topics_dest_group if topic in source_topics]
    if common_topics:
        print(
            "Error: Consumer Group ID: {groupid} is already "
            "subscribed to following topics: {topic}.\nPlease delete this "
            "topics from new group before re-running the "
            "command.".format(
                groupid=dest_groupid,
                topic=', '.join(common_topics),
            ),
            file=sys.stderr,
        )
        sys.exit(1)
    # Let's confirm what the user intends to do.
    if topics_dest_group:
        in_str = (
            "New Consumer Group: {dest_groupid} already "
            "exists.\nTopics subscribed to by the consumer groups are listed "
            "below:\n{source_groupid}: {source_group_topics}\n"
            "{dest_groupid}: {dest_group_topics}\nDo you intend to copy into"
            "existing consumer destination-group? (y/n)".format(
                source_groupid=source_groupid,
                source_group_topics=source_topics,
                dest_groupid=dest_groupid,
                dest_group_topics=topics_dest_group,
            )
        )
        prompt_user_input(in_str)


def create_offsets(zk, consumer_group, offsets):
    """Create path with offset value for each topic-partition of given consumer
    group.

    :param zk: Zookeeper client
    :param consumer_group: Consumer group id for given offsets
    :type consumer_group: int
    :param offsets: Offsets of all topic-partitions
    :type offsets: dict(topic, dict(partition, offset))
    """
    # Create new offsets
    for topic, partition_offsets in offsets.items():
        for partition, offset in partition_offsets.items():
            new_path = "/consumers/{groupid}/offsets/{topic}/{partition}".format(
                groupid=consumer_group,
                topic=topic,
                partition=partition,
            )
            try:
                zk.create(new_path, value=offset, makepath=True)
            except NodeExistsError:
                print(
                    "Error: Path {path} already exists. Please re-run the "
                    "command.".format(path=new_path),
                    file=sys.stderr,
                )
                raise


def fetch_offsets(zk, consumer_group, topics):
    """Fetch offsets for given topics of given consumer group.

    :param zk: Zookeeper client
    :param consumer_group: Consumer group id for given offsets
    :type consumer_group: int
    :rtype: dict(topic, dict(partition, offset))
    """
    source_offsets = defaultdict(dict)
    for topic, partitions in topics.items():
        for partition in partitions:
            offset, _ = zk.get(
                "/consumers/{groupid}/offsets/{topic}/{partition}".format(
                    groupid=consumer_group,
                    topic=topic,
                    partition=partition,
                )
            )
            source_offsets[topic][partition] = offset
    return source_offsets


def prompt_user_input(in_str):
    while True:
        answer = input(in_str + ' ')
        if answer == "n" or answer == "no":
            sys.exit(0)
        if answer == "y" or answer == "yes":
            return


def get_offset_topic_partition_count(kafka_config):
    """Given a kafka cluster configuration, return the number of partitions
    in the offset topic. It will raise an UnknownTopic exception if the topic
    cannot be found."""
    metadata = get_topic_partition_metadata(kafka_config.broker_list)
    if CONSUMER_OFFSET_TOPIC not in metadata:
        raise UnknownTopic("Consumer offset topic is missing.")
    return len(metadata[CONSUMER_OFFSET_TOPIC])


# The mapping of group information to partition for the __consumer_offsets
# topic is determinated by a hash of the group id, modulo the number of
# partitions in the topic.
# Kafka code (from main/scala/kafka/coordinator/GroupMetadataManager.scala):
#   def partitionFor(groupId: String): Int =
#       Utils.abs(groupId.hashCode) % groupMetadataTopicPartitionCount
# hashCode returns the hash of the string according to the Java default string
# hashing algorithm. The algorithm is implemented in the inner function
# java_string_hashcode.
def get_group_partition(group, partition_count):
    """Given a group name, return the partition number of the consumer offset
    topic containing the data associated to that group."""
    def java_string_hashcode(s):
        h = 0
        for c in s:
            h = (31 * h + ord(c)) & 0xFFFFFFFF
        return ((h + 0x80000000) & 0xFFFFFFFF) - 0x80000000
    return abs(java_string_hashcode(group)) % partition_count


def topic_offsets_for_timestamp(consumer, timestamp, topics):
    """Given an initialized KafkaConsumer, timestamp, and list of topics,
    looks up the offsets for the given topics by timestamp. The returned
    offset for each partition is the earliest offset whose timestamp is greater than or
    equal to the given timestamp in the corresponding partition.

    Arguments:
        consumer (KafkaConsumer): an initialized kafka-python consumer
        timestamp (int): Unix epoch milliseconds. Unit should be milliseconds
            since beginning of the epoch (midnight Jan 1, 1970 (UTC))
        topics (list): List of topics whose offsets are to be fetched.
    :returns:
        ``{TopicPartition: OffsetAndTimestamp}``: mapping from partition
        to the timestamp and offset of the first message with timestamp
        greater than or equal to the target timestamp.
        Returns ``{TopicPartition: None}`` for specific topic-partiitons if:
          1. Timestamps are not supported in messages
          2. No offsets in the partition after the given timestamp
          3. No data in the topic-partition
    :raises:
        ValueError: If the target timestamp is negative
        UnsupportedVersionError: If the broker does not support looking
            up the offsets by timestamp.
        KafkaTimeoutError: If fetch failed in request_timeout_ms
    """
    tp_timestamps = {}
    for topic in topics:
        topic_partitions = consumer_partitions_for_topic(consumer, topic)
        for tp in topic_partitions:
            tp_timestamps[tp] = timestamp
    return consumer.offsets_for_times(tp_timestamps)


def consumer_partitions_for_topic(consumer, topic):
    """Returns a list of all TopicPartitions for a given topic.

    Arguments:
        consumer: an initialized KafkaConsumer
        topic: a topic name to fetch TopicPartitions for

    :returns:
        list(TopicPartition): A list of TopicPartitions that belong to the given topic
    """
    topic_partitions = []
    partitions = consumer.partitions_for_topic(topic)
    if partitions is not None:
        for partition in partitions:
            topic_partitions.append(TopicPartition(topic, partition))
    else:
        logging.error(
            f"No partitions found for topic {topic}. Maybe it doesn't exist?",
        )
    return topic_partitions


def consumer_commit_for_times(consumer, partition_to_offset, atomic=False):
    """Commits offsets to Kafka using the given KafkaConsumer and offsets, a mapping
    of TopicPartition to Unix Epoch milliseconds timestamps.

    Arguments:
        consumer (KafkaConsumer): an initialized kafka-python consumer.
        partitions_to_offset (dict TopicPartition: OffsetAndTimestamp): Map of TopicPartition to OffsetAndTimestamp. Return value of offsets_for_times.
        atomic (bool): Flag to specify whether the commit should fail if offsets are not found for some
            TopicPartition: timestamp pairs.
    """
    no_offsets = set()
    for tp, offset in partition_to_offset.items():
        if offset is None:
            logging.error(
                "No offsets found for topic-partition {tp}. Either timestamps not supported"
                " for the topic {tp}, or no offsets found after timestamp specified, or there is no"
                " data in the topic-partition.".format(tp=tp),
            )
            no_offsets.add(tp)
    if atomic and len(no_offsets) > 0:
        logging.error(
            "Commit aborted; offsets were not found for timestamps in"
            " topics {}".format(",".join([str(tp) for tp in no_offsets])),
        )
        return

    offsets_metadata = {
        tp: OffsetAndMetadata(partition_to_offset[tp].offset, metadata=None)
        for tp in partition_to_offset.keys() if tp not in no_offsets
    }

    if len(offsets_metadata) != 0:
        consumer.commit(offsets_metadata)


class InvalidMessageException(Exception):
    pass


class KafkaAdminGroupReader:

    def __init__(self, kafka_config):
        self.log = logging.getLogger(__name__)
        self.admin_client = KafkaAdminClient(
            bootstrap_servers=kafka_config.broker_list,
        )

    def read_group(self, groupid):
        topics = set()
        group_offsets = self.admin_client.list_consumer_group_offsets(groupid)
        for tp in group_offsets.keys():
            topics.add(tp.topic)

        return list(topics)

    def read_groups(self, groupids=None, list_only=False):
        if groupids is None:
            groupids = self._list_groups()

        if list_only:
            return {groupid: [] for groupid in groupids}

        groups = {}
        for groupid in groupids:
            topics = self.read_group(groupid)
            groups[groupid] = topics

        return groups

    def _list_groups(self):
        groups_and_protocol_types = self.admin_client.list_consumer_groups()
        return [
            gpt[0]
            for gpt in groups_and_protocol_types
        ]


class KafkaGroupReader:

    def __init__(self, kafka_config):
        self.log = logging.getLogger(__name__)
        self.kafka_config = kafka_config
        self._kafka_groups = defaultdict(lambda: defaultdict(dict))
        self.active_partitions = {}
        self._finished = False

    def read_group(self, group_id):
        partition_count = get_offset_topic_partition_count(self.kafka_config)
        partition = get_group_partition(group_id, partition_count)
        return self.read_groups(partition)[group_id]

    def read_groups(self, partition=None, list_only=False):
        self.consumer = KafkaConsumer(
            group_id='offset_monitoring_consumer',
            bootstrap_servers=self.kafka_config.broker_list,
            auto_offset_reset='earliest',
            enable_auto_commit=False,
            consumer_timeout_ms=30000,
            fetch_max_wait_ms=2000,
            max_partition_fetch_bytes=10 * 1024 * 1024,  # 10MB
        )

        # Fetch metadata as partitions_for_topic only returns locally cached metadata
        # See https://github.com/dpkp/kafka-python/issues/1742
        self.consumer.topics()

        if partition is not None:
            self.active_partitions = {
                partition: TopicPartition(CONSUMER_OFFSET_TOPIC, partition),
            }
        else:
            self.active_partitions = {
                p: TopicPartition(CONSUMER_OFFSET_TOPIC, p)
                for p in self.consumer.partitions_for_topic(CONSUMER_OFFSET_TOPIC)
            }
        self.watermarks = self.get_current_watermarks(list(self.active_partitions.values()))
        # Active partitions are not empty. Remove the empty ones.
        self.active_partitions = {
            p: tp for p, tp in self.active_partitions.items()
            if tp.partition in self.watermarks and
            self.watermarks[tp.partition].highmark > 0 and
            self.watermarks[tp.partition].highmark > self.watermarks[tp.partition].lowmark
        }
        # Cannot consume if there are no active partitions
        if not self.active_partitions:
            return {}

        self.consumer.assign(list(self.active_partitions.values()))
        self.log.info("Consuming from %s", self.active_partitions)

        message_iterator = iter(self.consumer)

        while not self.finished():
            try:
                message = next(message_iterator)
            except StopIteration:
                continue
            # Stop when reaching the last message written to the
            # __consumer_offsets topic when KafkaGroupReader first started
            if message.offset >= self.watermarks[message.partition].highmark - 1:
                self.remove_partition_from_consumer(message.partition)
            self.process_consumer_offset_message(message)

        self._remove_unsubscribed_topics()

        return {
            group: topics.keys()
            for group, topics in self._kafka_groups.items()
            if topics
        }

    def _remove_unsubscribed_topics(self):
        for group, topics in list(self._kafka_groups.items()):
            for topic, partitions in list(topics.items()):
                # If offsets for all partitions are 0, consider the topic as unsubscribed
                if not any(partitions.values()):
                    del self._kafka_groups[group][topic]
                    self.log.info(f"Removed group {group} topic {topic} from list of groups")

    def remove_partition_from_consumer(self, partition):
        deleted = self.active_partitions.pop(partition)
        # Terminate if there are no more partitions to consume
        if not self.active_partitions:
            self.log.info("Completed reading from all partitions")
            self._finished = True
            return
        # Reassign the remaining partitions to the consumer while saving the
        # position
        positions = [
            (p, self.consumer.position(p))
            for p in self.active_partitions.values()
        ]
        self.consumer.assign(list(self.active_partitions.values()))
        for topic_partition, position in positions:
            self.consumer.seek(topic_partition, position)
        self.log.info(
            "Completed reading from %s. Remaining partitions: %s",
            deleted,
            self.active_partitions,
        )

    def parse_consumer_offset_message(self, message):
        key = message.key
        ((key_schema,), cur) = relative_unpack(b'>h', key, 0)
        if key_schema not in [0, 1]:
            raise InvalidMessageException()   # This is not an offset commit message
        (group, cur) = read_short_string(key, cur)
        (topic, cur) = read_short_string(key, cur)
        ((partition,), cur) = relative_unpack(b'>l', key, cur)
        if message.value:
            value = message.value
            ((value_schema,), cur) = relative_unpack(b'>h', value, 0)
            if value_schema not in [0, 1, 3]:
                raise InvalidMessageException()  # Unrecognized message value
            ((offset,), cur) = relative_unpack(b'>q', value, cur)
        else:
            offset = None  # Offset was deleted
        return group.decode(), topic.decode(), partition, offset

    def process_consumer_offset_message(self, message):
        try:
            group, topic, partition, offset = self.parse_consumer_offset_message(message)
        except InvalidMessageException:
            return

        if offset is not None:
            self._kafka_groups[group][topic][partition] = offset
            self.log.info(
                "Updated group {group} topic {topic} and updated offset in list of groups".format(
                    group=group,
                    topic=topic,
                ),
            )
        # TODO: check if we can ever find an offset commit message with message.value is None
        elif offset is None and group in self._kafka_groups and \
                topic in self._kafka_groups[group]:  # No offset means topic deletion
            del self._kafka_groups[group][topic]
            self.log.info(f"Removed group {group} topic {topic} from list of groups")

    def get_current_watermarks(self, partitions=None):
        client = KafkaToolClient(self.kafka_config.broker_list)
        client.load_metadata_for_topics(CONSUMER_OFFSET_TOPIC)
        offsets = get_topics_watermarks(
            client,
            [CONSUMER_OFFSET_TOPIC],
        )
        partitions_set = {tp.partition for tp in partitions} if partitions else None
        return {part: offset for part, offset
                in offsets[CONSUMER_OFFSET_TOPIC].items()
                if offset.highmark > offset.lowmark and
                (partitions is None or part in partitions_set)}

    def finished(self):
        return self._finished


def get_kafka_group_reader(cluster_config, use_admin_client=False):
    if use_admin_client:
        return KafkaAdminGroupReader(cluster_config)
    else:
        return KafkaGroupReader(cluster_config)



# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import importlib
import inspect
import os
import sys


def get_module(module_full_name):
    if ':' in module_full_name:
        path, module_name = module_full_name.rsplit(':', 1)
        if not os.path.isdir(path):
            print(f"{path} is not a valid directory", file=sys.stderr)
            sys.exit(1)
        sys.path.append(path)
        return importlib.import_module(module_name)
    else:
        return importlib.import_module(module_full_name)


def child_class(class_types, base_class):
    """
    Find the child-most class of `base_class`.

    Examples:
        class A:
            pass

        class B(A):
            pass

        class C(B):
            pass

        child_class([A, B, C], A) = C
    """
    subclasses = set()
    for class_type in class_types:
        if class_type is base_class:
            continue
        if issubclass(class_type, base_class):
            subclasses.add(class_type)

    if len(subclasses) == 0:
        return None
    elif len(subclasses) == 1:
        return subclasses.pop()
    else:
        # If more than one class is a subclass of `base_class`
        # It is possible that one or more classes are subclasses of another
        # class (see example above).
        # Recursively find the child-most class. Break ties by returning any
        # child-most class.
        for c in subclasses:
            child = child_class(subclasses, c)
            if child is not None:
                return child
        return subclasses.pop()


def dynamic_import(module_full_name, base_class):
    module = get_module(module_full_name)
    class_types = [
        class_type
        for _, class_type in inspect.getmembers(module, inspect.isclass)
    ]
    return child_class(class_types, base_class)



# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import logging
import sys

from .command import ClusterManagerCmd


class PreferredReplicaElectionCmd(ClusterManagerCmd):

    def __init__(self):
        super().__init__()
        self.log = logging.getLogger(self.__class__.__name__)

    def build_subparser(self, subparsers):
        subparser = subparsers.add_parser(
            'preferred-replica-election',
            description='Generates json files for kafka-preferred-replica-election',
            help='This command is used to help generate kafka-preferred-replica-election'
                 ' json files based on input from kafka-check.'
        )
        subparser.add_argument(
            '--topic-partition-filter',
            type=str,
            required=True,
            help='Path to file containing a colon separated list of topic partitions to work on.'
                 ' This is useful if you only want to operate on a limited set of topic partitions'
                 ' Observe that kafka-check outputs colon separated lists, after ommitting the first $n'
                 ' lines e.g. kafka-check offline | tail -n+$N',
        )
        return subparser

    def run_command(self, cluster_topology, cluster_balancer):
        filter_set = self.get_topic_filter()
        output = {'partitions': []}
        # Validate the filter_set to check that all of the partitions provided are part of the cluster_topology
        for t_p in filter_set:
            if t_p not in cluster_topology.assignment:
                self.log.error(f'Topic partition {t_p} from filter not in cluster topology')
                sys.exit(1)
            topic, partition = t_p[0], t_p[1]
            output['partitions'].append({'topic': topic, 'partition': partition})

        self.log.info('Preferred replica assignment json generated. Run kafka-preferred-replica-election using this json:')
        print(json.dumps(output))

        if self.args.proposed_plan_file:
            self.write_json_plan(output, self.args.proposed_plan_file)
            self.log.info(f'Saved plan to file {self.args.proposed_plan_file}')



# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import sys
from collections import defaultdict
from collections import OrderedDict
from datetime import datetime
from typing import DefaultDict

import pytz
from kafka import KafkaConsumer
from kafka.structs import TopicPartition

from .offset_manager import OffsetManagerBase
from kafka_utils.kafka_consumer_manager.util import consumer_commit_for_times
from kafka_utils.kafka_consumer_manager.util import topic_offsets_for_timestamp


class OffsetSetTimestamp(OffsetManagerBase):

    new_offsets_dict: DefaultDict[str, dict[int, int]] = defaultdict(dict)

    @classmethod
    def topics_dict(cls, string):
        try:
            topic, partition_timestamp = string.rsplit(".", 1)
            partition, timestamp = partition_timestamp.split("=", 1)
            cls.new_offsets_dict[topic][int(partition)] = int(timestamp)
        except ValueError:
            print(
                "Error: Badly formatted input, please re-run command "
                "with --help option.",
                file=sys.stderr,
            )
            sys.exit(1)

    @classmethod
    def setup_subparser(cls, subparsers):
        parser_offset_set = subparsers.add_parser(
            "offset_set_timestamp",
            description="Sets the offset(s) associated with the given timestamp for"
            " the given TopicPartitions for the consumer group. A single timestamp can"
            " also be provided to set offsets for all topic partitions the consumer group"
            " has offsets for. Note: all times are displayed in US/Pacific Standard Time.",
            add_help=False,
        )
        parser_offset_set.add_argument(
            "-h", "--help", action="help",
            help="Show this help message and exit.",
        )
        parser_offset_set.add_argument(
            "groupid",
            help="Consumer Group ID whose consumer offsets shall be modified.",
        )
        parser_offset_set.add_argument(
            "--timestamp",
            type=int,
            help="The Unix Epoch timestamp, in milliseconds, to set all offsets"
            " of the consumer group to.",
        )
        parser_offset_set.add_argument(
            "newoffsets", nargs="*", metavar=("<topic>.<partition>=<timestamp>"),
            type=cls.topics_dict,
            help="Tuple containing the Kafka topic, partition and"
            " the intended timestamp to use for setting offsets.",
        )
        parser_offset_set.add_argument(
            "--atomic",
            action="store_true",
            help="Don't commit any offsets if any TopicPartition: timestamp mapping"
            " has no offsets found. By default, any offsets that can be committed are"
            " committed.",
        )
        parser_offset_set.set_defaults(command=cls.run)

    @classmethod
    def validate_args(cls, args):
        if args.timestamp and len(cls.new_offsets_dict) > 0:
            print("Error: cannot supply both a timestamp and list of topic.partition=timestamp.")
            print("Re-run with --help for more information.")
            return False
        return True

    @classmethod
    def run(cls, args, cluster_config):

        if not cls.validate_args(args):
            return

        # Set up the KafkaConsumer
        consumer = KafkaConsumer(
            bootstrap_servers=cluster_config.broker_list,
            client_id="kafka-consumer-manager",
            group_id=args.groupid,
            enable_auto_commit=False,
        )

        tp_timestamps = {}
        # Get topics that the consumer group has committed offsets for in the past
        topics = cls.get_topics_from_consumer_group_id(cluster_config, args.groupid)

        if args.timestamp and len(cls.new_offsets_dict) == 0:
            # Committing for using --timestamp for all topics consumer group follows
            if len(topics) == 0:
                return
            partition_to_offset = topic_offsets_for_timestamp(consumer, args.timestamp, topics)
            for tp in partition_to_offset.keys():
                tp_timestamps[tp] = args.timestamp
        else:
            # Committing via a list of topic.partition=timestamp
            for topic in cls.new_offsets_dict.keys():
                for partition in cls.new_offsets_dict[topic].keys():
                    tp_timestamps[TopicPartition(topic, partition)] = cls.new_offsets_dict[topic][partition]
            partition_to_offset = consumer.offsets_for_times(tp_timestamps)

        cls.print_offsets(partition_to_offset, tp_timestamps)

        # Display warning if consumer is attempting to commit offsets for a topic they
        # have never committed offsets for.
        for tp in partition_to_offset.keys():
            if tp.topic not in topics:
                print(
                    "WARNING: Consumer group {cg} has never committed offsets for topic {topic}.".format(
                        cg=args.groupid,
                        topic=tp.topic,
                    )
                )

        if cls.prompt_user():
            consumer_commit_for_times(consumer, partition_to_offset, atomic=args.atomic)
        else:
            print("Offsets not committed")

    @classmethod
    def prompt_user(cls):
        print("The above TopicPartitions and offsets will attempt to be committed.")
        print("Continue? (y/n)")

        return input() == "y"

    @classmethod
    def print_offsets(cls, partition_to_offset, tp_timestamps):
        topics = {}
        for tp, offset_timestamp in partition_to_offset.items():
            if tp.topic not in topics:
                topics[tp.topic] = {}
            topics[tp.topic][tp.partition] = offset_timestamp
        topics = OrderedDict(sorted(topics.items(), key=lambda k: k[0]))
        for topic in topics.keys():
            topics[topic] = OrderedDict(sorted(topics[topic].items(), key=lambda k: k[0]))
            print(f"Topic Name: {topic}")
            for partition, offset_timestamp in topics[topic].items():
                print(
                    f"\tPartition ID: {partition}",
                )
                offset = "not found"
                timestamp = tp_timestamps[TopicPartition(topic, partition)]
                if offset_timestamp is not None:
                    offset = offset_timestamp.offset
                    timestamp = offset_timestamp.timestamp
                date = datetime.fromtimestamp(
                    timestamp / 1000.0,
                    tz=pytz.timezone("US/Pacific"),
                ).strftime("%Y-%m-%d %H:%M:%S %Z")
                print(
                    "\t\tTimestamp: {timestamp} ({date})".format(
                        timestamp=timestamp,
                        date=date,
                    ),
                )
                print(f"\t\tOffset: {offset}")



# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import logging
import sys

import humanfriendly

from kafka_utils.util import config
from kafka_utils.util.zookeeper import ZK


LEADER_THROTTLE_RATE_CONFIGURATION = "leader.replication.throttled.rate"
FOLLOWER_THROTTLE_RATE_CONFIGURATION = "follower.replication.throttled.rate"


def parse_opts():
    parser = argparse.ArgumentParser(
        description='Manually applies throttle values for brokers in a Kafka cluster',
    )
    parser.add_argument(
        '--cluster-type',
        '-t',
        required=True,
        help='cluster type, e.g. "standard"',
    )
    parser.add_argument(
        '--cluster-name',
        '-c',
        help='cluster name, e.g. "uswest1-devc" (defaults to local cluster)',
    )
    parser.add_argument(
        '--discovery-base-path',
        dest='discovery_base_path',
        type=str,
        help='Path of the directory containing the <cluster_type>.yaml config',
    )
    parser.add_argument(
        '--leader-throttle',
        type=int,
        required=False,
        help='value (B/s) of the leader throttle to apply for brokers',
    )
    parser.add_argument(
        '--follower-throttle',
        type=int,
        required=False,
        help='value (B/s) of the leader throttle to apply for brokers',
    )
    parser.add_argument(
        '--clear',
        help='if set, throttles will be cleared instead of being set',
        action='store_true',
    )
    parser.add_argument(
        '--read-only',
        help='if set, only reads throttles and exits',
        action='store_true',
    )
    parser.add_argument(
        '-v',
        '--verbose',
        help='print verbose execution information. Default: %(default)s',
        action="store_true",
    )
    return parser.parse_args()


def validate_opts(opts):
    """
    Basic option validation.

    True if the options are valid, False otherwise.

    :opts : the command line options
    :returns : bool
    """
    if opts.read_only:
        return True

    if opts.clear:
        if opts.leader_throttle is not None:
            print("Error: --clear cannot be used with --leader-throttle")
            return False
        if opts.follower_throttle is not None:
            print("Error: --clear cannot be used with --follower-throttle")
            return False
    else:
        if opts.leader_throttle < 0:
            print("Error: --leader-throttle must be >= 0")
            return False
        if opts.follower_throttle < 0:
            print("Error: --follower-throttle must be >= 0")
            return False
    return True


def human_throttle(throttle):
    if throttle is None:
        return "N/A"

    return humanfriendly.format_size(int(throttle), binary=True) + "/s"


def print_throttles(zk, brokers):
    """
    Print the current replication throttles.

    Throttles are written in B/s, and as a human readable format.

    :zk : a ZK client
    :brokers : a collection of broker ids
    """

    print("Reading current replication throttles.")

    broker_throttles = read_throttles(zk, brokers)

    for broker_id, (leader_throttle, follower_throttle) in broker_throttles.items():
        print(
            "\tBroker ID: {broker_id} - Leader: {leader_throttle} ({leader_human}) - Follower: {follower_throttle} ({follower_human})".format(
                broker_id=broker_id,
                leader_throttle=leader_throttle,
                follower_throttle=follower_throttle,
                leader_human=human_throttle(leader_throttle),
                follower_human=human_throttle(follower_throttle),
            )
        )


def read_throttles(zk, brokers):
    """
    Read leader/follower replication throttles for the given brokers

    :zk : a ZK client
    :brokers : a collection of broker ids
    :returns : a mapping of broker id -> (leader throttle, follower throttle) pairs.
                    throttles are in B/s, None if no value is configured
    """
    throttles = {}

    for broker_id in brokers:
        config = zk.get_broker_config(broker_id).get('config', {})

        leader_throttle = config.get(LEADER_THROTTLE_RATE_CONFIGURATION)
        follower_throttle = config.get(FOLLOWER_THROTTLE_RATE_CONFIGURATION)

        throttles[broker_id] = ((leader_throttle, follower_throttle))

    return throttles


def apply_throttles(zk, brokers, leader_throttle, follower_throttle):
    """
    Apply new leader/follower replication throttles to the given brokers

    :zk : a ZK client
    :brokers : a collection of broker ids
    :leader_throttle : new leader replication throttle (in B/s)
    :follower_throttle : new follower replication throttle (in B/s)
    """
    for broker_id in brokers:
        write_throttle(zk, broker_id, str(leader_throttle), str(follower_throttle))


def clear_throttles(zk, brokers):
    """
    Clear replication throttles for the given brokers.

    :zk : a ZK client
    :brokers : a collection of broker ids
    """
    for broker_id in brokers:
        write_throttle(zk, broker_id, None, None)


def write_throttle(zk, broker_id, leader_throttle, follower_throttle):
    """
    Write leader/follower replication throttle rates for a given broker.

    If a given replication throttle is None, the configuration will be cleared
    from the broker instead of applying a new one.

    More details can be found in:
        https://kafka.apache.org/documentation/#rep-throttle

    :zk : a ZK client
    :broker_id : broker to change replication throttles for
    :leader_throttle : new leader replication throttle (in B/s) or None
    :follower_throttle : new follower replication throttle (in B/s) or None
    """
    config = zk.get_broker_config(broker_id).get('config', {})

    if leader_throttle is not None:
        config[LEADER_THROTTLE_RATE_CONFIGURATION] = leader_throttle
    else:
        config.pop(LEADER_THROTTLE_RATE_CONFIGURATION, None)

    if follower_throttle is not None:
        config[FOLLOWER_THROTTLE_RATE_CONFIGURATION] = follower_throttle
    else:
        config.pop(FOLLOWER_THROTTLE_RATE_CONFIGURATION, None)

    zk.set_broker_config(broker_id, {'config': config})


def run():
    opts = parse_opts()
    if opts.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARN)

    if not validate_opts(opts):
        sys.exit(1)

    cluster_config = config.get_cluster_config(
        opts.cluster_type,
        opts.cluster_name,
        opts.discovery_base_path,
    )

    with ZK(cluster_config) as zk:
        brokers = zk.get_brokers(names_only=True)

        print_throttles(zk, brokers)

        if opts.read_only:
            return

        print("Applying new replication throttles")

        if not opts.clear:
            apply_throttles(
                zk,
                brokers,
                opts.leader_throttle,
                opts.follower_throttle,
            )
        else:
            clear_throttles(zk, brokers)

        print("New replication throttles applied.")
        print_throttles(zk, brokers)

    if not opts.clear:
        print("NOTE: Do not forget to --clear throttles once the reassignment plan completes.")



# Copyright 2018 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import itertools
from typing import Any

from kafka.structs import PartitionMetadata
from typing_extensions import TypedDict

from kafka_utils.kafka_check import status_code
from kafka_utils.kafka_check.commands.command import KafkaCheckCmd
from kafka_utils.kafka_check.commands.min_isr import get_min_isr
from kafka_utils.util.metadata import get_topic_partition_metadata
from kafka_utils.util.zookeeper import ZK


class ReplicationFactorCmd(KafkaCheckCmd):

    def build_subparser(self, subparsers: Any) -> Any:
        subparser = subparsers.add_parser(
            'replication_factor',
            description='Check replication factor settings for each topic in the cluster.',
            help='This command will check replication factor each topic in the cluster '
                 'and compare it with min.isr settings in Zookeeper or default min.isr param '
                 'if it is specified and there is no settings in Zookeeper for a topic.',
        )
        subparser.add_argument(
            '--default-min-isr',
            type=int,
            default=1,
            help='Default min.isr value for cases without settings in Zookeeper '
            'for some topics. Default: %(default)s',
        )
        return subparser

    def run_command(self) -> tuple[int, dict[str, Any]]:
        """Replication factor command, checks replication factor settings and compare it with
        min.isr in the cluster."""
        topics = get_topic_partition_metadata(self.cluster_config.broker_list)

        topics_with_wrong_rf = _find_topics_with_wrong_rp(
            topics,
            self.zk,
            self.args.default_min_isr,
        )

        errcode = status_code.OK if not topics_with_wrong_rf else status_code.CRITICAL
        out = _prepare_output(topics_with_wrong_rf, self.args.verbose, self.args.head)
        return errcode, out


class TopicDict(TypedDict):
    replication_factor: int
    min_isr: int
    topic: str


def _find_topics_with_wrong_rp(topics: dict[str, list[PartitionMetadata]], zk: ZK, default_min_isr: int) -> list[TopicDict]:
    """Returns topics with wrong replication factor."""
    topics_with_wrong_rf: list[TopicDict] = []

    for topic_name, partitions in topics.items():
        min_isr = get_min_isr(zk, topic_name) or default_min_isr
        replication_factor = len(partitions[0].replicas)

        if replication_factor >= min_isr + 1:
            continue

        topics_with_wrong_rf.append({
            'replication_factor': replication_factor,
            'min_isr': min_isr,
            'topic': topic_name,
        })

    return topics_with_wrong_rf


def _prepare_output(topics_with_wrong_rf: list[TopicDict], verbose: bool, head_limit: int) -> dict[str, Any]:
    """Returns dict with 'raw' and 'message' keys filled."""
    out: dict[str, Any] = {}
    topics_count = len(topics_with_wrong_rf)
    out['raw'] = {
        'topics_with_wrong_replication_factor_count': topics_count,
    }
    if head_limit != -1:
        topics_with_wrong_rf = list(itertools.islice(topics_with_wrong_rf, head_limit))

    if topics_count == 0:
        out['message'] = 'All topics have proper replication factor.'
    else:
        out['message'] = (
            "{} topic(s) have replication factor lower than specified min ISR + 1."
        ).format(topics_count)

        if verbose:
            lines = (
                "replication_factor={replication_factor} is lower than min_isr={min_isr} + 1 for {topic}"
                .format(
                    min_isr=topic['min_isr'],
                    topic=topic['topic'],
                    replication_factor=topic['replication_factor'],
                )
                for topic in topics_with_wrong_rf
            )
            if head_limit != -1:
                title = f"Top {head_limit} topics:\n"
            else:
                title = "Topics:\n"

            out['verbose'] = title + "\n".join(lines)

    if verbose:
        out['raw']['topics'] = topics_with_wrong_rf

    return out



# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from kafka.errors import UnknownMemberIdError

from .offset_manager import OffsetWriter
from kafka_utils.util.client import KafkaToolClient
from kafka_utils.util.offsets import get_current_consumer_offsets
from kafka_utils.util.offsets import nullify_offsets
from kafka_utils.util.offsets import set_consumer_offsets


class UnsubscribeTopics(OffsetWriter):

    @classmethod
    def setup_subparser(cls, subparsers):
        parser_unsubscribe_topics = subparsers.add_parser(
            "unsubscribe_topics",
            description="Delete topics and partitions by consumer group. This "
            "tool shall delete all offset metadata from Zookeeper.",
            add_help=False
        )
        parser_unsubscribe_topics.add_argument(
            "-h", "--help", action="help",
            help="Show this help message and exit."
        )
        parser_unsubscribe_topics.add_argument(
            'groupid',
            help="Consumer Group IDs whose metadata shall be deleted."
        )
        parser_unsubscribe_topics.add_argument(
            '--topic',
            help="Topic whose metadata shall be deleted. If either --topic or"
            "--topics are NOT specified, all topics that the consumer is"
            "subscribed to, shall be deleted."
        )
        parser_unsubscribe_topics.add_argument(
            '--partitions', nargs='+', type=int,
            help="List of partitions whose metadata shall be deleted. If no "
            "partitions are specified, all partitions within the topic shall "
            "be deleted. Only works with --topic, NOT with --topics."
        )
        parser_unsubscribe_topics.add_argument(
            '--topics', nargs='+',
            help="Topics whose metadata shall be deleted. If either --topic or"
            "--topics are NOT specified, all topics that the consumer is"
            "subscribed to, shall be deleted."
        )
        parser_unsubscribe_topics.set_defaults(command=cls.run)

    @classmethod
    def run(cls, args, cluster_config):
        # Setup the Kafka client
        client = KafkaToolClient(cluster_config.broker_list)
        client.load_metadata_for_topics()

        # if topic is not None topics_dict will contain only info about that
        # topic, otherwise it will contain info about all topics for the group
        topics_dict = cls.preprocess_args(
            args.groupid,
            args.topic,
            args.partitions,
            cluster_config,
            client,
            topics=args.topics,
            use_admin_client=args.use_admin_client,
        )

        topics = args.topics if args.topics else ([args.topic] if args.topic else [])

        unsubscriber = KafkaUnsubscriber(client)
        unsubscriber.unsubscribe_topics(
            args.groupid,
            topics,
            args.partitions,
            topics_dict,
        )


class TopicUnsubscriber:
    """Base class used to unsubscribe consumer groups from topic partitions."""

    def unsubscribe_topics(self, group, topics, partitions, topics_dict):
        # If a single topic and partitions are both specified,
        # then unsubscribe the group from the individual partitions

        # If only the topics are specified, the unsubscribe the group
        # from all the partitions of the topics.

        # If neither the topic nor partitions are specified, then
        # unsubscribe the group from all of the topics and partitions
        # that are found in the topics_dict that was preprocessed.
        if topics and len(topics) == 1 and partitions:
            self.unsubscribe_partitions(group, topics[0], partitions)
        elif topics:
            for topic in topics:
                try:
                    self.delete_topic(group, topic)
                except UnknownMemberIdError:
                    print(
                        "Error: Unable to unsubscribe group '{group_name}' from topic '{topic_name}'. \
                            Please ensure all consumers in this consumer group are stopped before \
                            trying to unsubscribe a consumer group with offsets stored in \
                            Kafka.".format(group_name=group, topic_name=topic),
                        file=sys.stderr,
                    )
                    raise

        else:
            for topic, partitions in topics_dict.items():
                self.delete_topic(group, topic)

    def unsubscribe_partitions(self, group, topic, partitions):
        raise NotImplementedError()

    def delete_topic(self, group, topic):
        raise NotImplementedError()


class KafkaUnsubscriber(TopicUnsubscriber):
    """Class used to unsubscribe consumer groups."""

    def __init__(self, client):
        self.client = client

    def unsubscribe_partitions(self, group, topic, partitions):
        offsets = {
            topic: {
                partition: 0
                for partition in partitions
            }
        }
        set_consumer_offsets(
            self.client,
            group,
            nullify_offsets(offsets),
        )

    def delete_topic(self, group, topic):
        offsets = get_current_consumer_offsets(
            self.client,
            group,
            [topic],
        )
        set_consumer_offsets(
            self.client,
            group,
            nullify_offsets(offsets),
        )



# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import sys

from .offset_manager import OffsetManagerBase
from kafka_utils.util.client import KafkaToolClient
from kafka_utils.util.offsets import get_current_consumer_offsets
from kafka_utils.util.offsets import nullify_offsets
from kafka_utils.util.offsets import set_consumer_offsets


class RenameGroup(OffsetManagerBase):

    @classmethod
    def setup_subparser(cls, subparsers):
        parser_rename_group = subparsers.add_parser(
            "rename_group",
            description="Rename specified consumer group ID to a new name. "
            "This tool shall migrate all offset metadata in Zookeeper.",
            add_help=False
        )
        parser_rename_group.add_argument(
            "-h", "--help", action="help",
            help="Show this help message and exit."
        )
        parser_rename_group.add_argument(
            'old_groupid',
            help="Consumer Group ID to be renamed."
        )
        parser_rename_group.add_argument(
            'new_groupid',
            help="New name for the consumer group ID."
        )
        parser_rename_group.set_defaults(command=cls.run)

    @classmethod
    def run(cls, args, cluster_config):
        if args.old_groupid == args.new_groupid:
            print(
                "Error: Old group ID and new group ID are the same.",
                file=sys.stderr,
            )
            sys.exit(1)
        # Setup the Kafka client
        client = KafkaToolClient(cluster_config.broker_list)
        client.load_metadata_for_topics()

        topics_dict = cls.preprocess_args(
            groupid=args.old_groupid,
            topic=None,
            partitions=None,
            cluster_config=cluster_config,
            client=client,
            use_admin_client=args.use_admin_client,
        )
        cls.rename_group(
            client,
            args.old_groupid,
            args.new_groupid,
            topics_dict,
        )

    @classmethod
    def rename_group(
        cls,
        client,
        old_groupid,
        new_groupid,
        topics,
    ):
        copied_offsets = get_current_consumer_offsets(
            client,
            old_groupid,
            topics,
        )
        set_consumer_offsets(client, new_groupid, copied_offsets)
        set_consumer_offsets(
            client,
            old_groupid,
            nullify_offsets(topics),
        )



# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import sys

import humanfriendly

from .command import ClusterManagerCmd
from .command import DEFAULT_MAX_MOVEMENT_SIZE
from kafka_utils.util import positive_float
from kafka_utils.util import positive_int
from kafka_utils.util.validation import assignment_to_plan
from kafka_utils.util.validation import validate_plan


DEFAULT_MAX_PARTITION_MOVEMENTS = 1
DEFAULT_MAX_LEADER_CHANGES = 5


class DecommissionCmd(ClusterManagerCmd):

    def __init__(self):
        super().__init__()
        self.log = logging.getLogger(self.__class__.__name__)

    def build_subparser(self, subparsers):
        subparser = subparsers.add_parser(
            'decommission',
            description='Decommission one or more brokers of the cluster.',
            help='This command is used to move all the replicas assigned to given '
            'brokers and redistribute them across all the other brokers while '
            'trying to keep the cluster balanced.',
        )
        subparser.add_argument(
            'broker_ids',
            nargs='+',
            type=int,
            help='Broker ids of the brokers to decommission.',
        )
        subparser.add_argument(
            '--max-partition-movements',
            type=positive_int,
            default=DEFAULT_MAX_PARTITION_MOVEMENTS,
            help='Maximum number of partition-movements in final set of actions.'
                 ' DEFAULT: %(default)s. RECOMMENDATION: Should be at least max '
                 'replication-factor across the cluster.',
        )
        subparser.add_argument(
            '--max-leader-changes',
            type=positive_int,
            default=DEFAULT_MAX_LEADER_CHANGES,
            help='Maximum number of actions with leader-only changes.'
                 ' DEFAULT: %(default)s',
        )

        movement_size_group = subparser.add_mutually_exclusive_group()
        movement_size_group.add_argument(
            '--max-movement-size',
            type=positive_float,
            default=DEFAULT_MAX_MOVEMENT_SIZE,
            help='Maximum total size of the partitions moved in the final set'
                 ' of actions. Since each PartitionMeasurer implementation'
                 ' defines its own notion of size, the size unit to use will'
                 ' depend on  the selected PartitionMeasurer implementation.'
                 ' DEFAULT: No limit.'
                 ' RECOMMENDATION: Should be at least the maximum partition-size'
                 ' on the brokers to decommission, ideally a little larger.',
        )
        movement_size_group.add_argument(
            '--auto-max-movement-size',
            action='store_true',
            help='Set max-movement-size to the size of the largest partition'
                 ' in the set of brokers to be decommissioned.',
        )
        subparser.add_argument(
            '--force-progress',
            action='store_true',
            help='If set, ensures that at minimum one partition is moved regardless'
                 ' of --max-movement-size set.',
        )
        return subparser

    def run_command(self, cluster_topology, cluster_balancer):
        # If the max_movement_size is still default, then the user did not input a value for it
        if self.args.force_progress and self.args.max_movement_size == DEFAULT_MAX_MOVEMENT_SIZE:
            self.log.error(
                '--force-progress must be used with --max-movement-size',
            )
            sys.exit(1)
        # Obtain the largest partition in the set of partitions we will move
        partitions_to_move = set()
        for broker in self.args.broker_ids:
            partitions_to_move.update(cluster_topology.brokers[broker].partitions)

        largest_size = max(
            [partition.size for partition in partitions_to_move] or (1, )
        )

        smallest_size = min(
            [partition.size for partition in partitions_to_move] or (0, )
        )

        if self.args.auto_max_movement_size:
            self.args.max_movement_size = largest_size
            self.log.info(
                'Auto-max-movement-size: using {human_max_movement_size} ({max_movement_size}) as'
                ' max-movement-size.'.format(
                    human_max_movement_size=humanfriendly.format_size(self.args.max_movement_size),
                    max_movement_size=self.args.max_movement_size,
                )
            )

        if self.args.max_movement_size and self.args.max_movement_size < largest_size:
            if not self.args.force_progress:
                self.log.error(
                    'Max partition movement size is only {max_movement_size},'
                    ' but remaining partitions to move range from {smallest_size} to'
                    ' {largest_size}. The decommission will not make progress'.format(
                        max_movement_size=humanfriendly.format_size(self.args.max_movement_size),
                        smallest_size=humanfriendly.format_size(smallest_size),
                        largest_size=humanfriendly.format_size(largest_size),
                    )
                )
                sys.exit(1)
            else:
                self.log.warning(
                    'Max partition movement size is only {max_movement_size},'
                    ' but remaining partitions to move range from {smallest_size} to'
                    ' {largest_size}. The decommission may be slower than expected'.format(
                        max_movement_size=humanfriendly.format_size(self.args.max_movement_size),
                        smallest_size=humanfriendly.format_size(smallest_size),
                        largest_size=humanfriendly.format_size(largest_size),
                    )
                )

        base_assignment = cluster_topology.assignment

        cluster_balancer.decommission_brokers(self.args.broker_ids)

        if not validate_plan(
            assignment_to_plan(cluster_topology.assignment),
            assignment_to_plan(base_assignment),
        ):
            self.log.error('Invalid assignment %s.', cluster_topology.assignment)
            print(
                f'Invalid assignment: {cluster_topology.assignment}',
                file=sys.stderr,
            )
            sys.exit(1)

        # Reduce the proposed assignment based on max_partition_movements
        # and max_leader_changes
        reduced_assignment = self.get_reduced_assignment(
            base_assignment,
            cluster_topology,
            self.args.max_partition_movements,
            self.args.max_leader_changes,
            max_movement_size=self.args.max_movement_size,
            force_progress=self.args.force_progress,
        )
        if reduced_assignment:
            self.process_assignment(reduced_assignment)
        else:
            msg_str = "Cluster already balanced. No more replicas in decommissioned brokers."
            self.log.info(msg_str)
            print(msg_str)



# Copyright 2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import itertools
import sys
from typing import Any

from kafka_utils.kafka_check import status_code
from kafka_utils.kafka_check.commands.command import KafkaCheckCmd
from kafka_utils.util.metadata import get_topic_partition_with_error
from kafka_utils.util.metadata import LEADER_NOT_AVAILABLE_ERROR


class OfflineCmd(KafkaCheckCmd):

    def build_subparser(self, subparsers: Any) -> Any:
        subparser = subparsers.add_parser(
            'offline',
            description='Check offline partitions on the specified broker',
            help='This subcommand will fail if there are any offline partitions '
            'in the cluster.'
        )

        return subparser

    def run_command(self) -> tuple[int, dict[str, Any]]:
        """Checks the number of offline partitions"""
        offline = get_topic_partition_with_error(
            self.cluster_config,
            LEADER_NOT_AVAILABLE_ERROR,
        )

        errcode = status_code.OK if not offline else status_code.CRITICAL
        out = _prepare_output(offline, self.args.verbose, self.args.head)
        return errcode, out


def _prepare_output(partitions: list[tuple[str, int]], verbose: bool, head_limit: int) -> dict[str, Any]:
    """Returns dict with 'raw' and 'message' keys filled."""
    out: dict[str, Any] = {}
    partitions_count = len(partitions)
    out['raw'] = {
        'offline_count': partitions_count,
    }

    if head_limit != -1:
        partitions = list(itertools.islice(partitions, head_limit))

    if partitions_count == 0:
        out['message'] = 'No offline partitions.'
    else:
        out['message'] = f"{partitions_count} offline partitions."
        if verbose:
            lines = (
                f'{topic}:{partition}'
                for (topic, partition) in partitions
            )
            title = f"Top {head_limit} partitions:\n" if head_limit != -1 else "Partitions:\n"
            out['verbose'] = title + "\n".join(lines)
        else:
            cmdline = sys.argv[:]
            cmdline.insert(1, '-v')
            out['message'] += '\nTo see all offline partitions run: ' + ' '.join(cmdline)

    if verbose:
        out['raw']['partitions'] = [
            {'topic': topic, 'partition': partition}
            for (topic, partition) in partitions
        ]

    return out
